# coding: utf8

from django.core.urlresolvers import reverse

from .models import Account, Transaction

from django.contrib.auth import get_user_model
create_user = get_user_model().objects.create_user

from django import test


class BaseTestCase(test.TestCase):
    """
    The base test case, providing convenience methods for various testing
    related tasks.
    """

    def setUp(self):
        self.client = test.client.Client()

    def check_template(self, url_name, templated_path,
                       method='get', **kwargs):
        """
        Checks the given template was used for the given url.
        """

        response = getattr(self, method)(
            url_name=url_name, url_kwargs=kwargs
        )
        self.assertTemplateUsed(response, templated_path)

    def get_url(self, name, **kwargs):
        """ Returns the url for the given name, args and kwargs. """

        return reverse(name, kwargs=kwargs)

    def get(self, url_name=None, url=None, url_kwargs={}, data={}):
        """
        Returns the response for a POST request on the url with the given name,
        args and kwargs.
        """

        if not url:
            url = self.get_url(url_name, **url_kwargs)

        return self.client.get(url, data=data)

    def post(self, url_name=None, url=None, url_kwargs={}, data={}):
        """
        Returns the response for a POST request on the url with the given name,
        args and kwargs.
        """

        if not url:
            url = self.get_url(url_name, **url_kwargs)

        return self.client.post(url, data=data)


class TestAccount(BaseTestCase):
    email1 = 'test1@example.com'
    password1 = 'test'

    email2 = 'test2@example.com'
    password2 = 'test'

    def setUp(self):
        self.user1 = create_user(self.email1, self.password1)
        self.user1_source = Account.GetPrimarySourceAccount(self.user1)
        self.user1_destination = Account.GetPrimaryDestinationAccount(
            self.user1
        )

        self.user2 = create_user(self.email2, self.password2)
        self.user2_source = Account.GetPrimarySourceAccount(self.user2)
        self.user2_destination = Account.GetPrimaryDestinationAccount(
            self.user2
        )

    def test_new_user_destination_account_creation(self):
        user = create_user('bob@dundermifflin.com', 'test')

        try:
            self.assertNotEqual(
                Account.GetPrimaryDestinationAccount(user), None
            )
        except Account.DoesNotExist:
            self.fail('New user destination account creation failed.')

    def test_new_user_source_account_creation(self):
        user = create_user('samantha@dundermifflin.com', 'test')

        try:
            self.assertNotEqual(
                Account.GetPrimarySourceAccount(user), None
            )
        except Account.DoesNotExist:
            self.fail('New user source account creation failed.')

    def test_new_user_balance(self):
        user = create_user('hope@dundermifflin.com', 'test')

        self.assertEqual(
            Account.GetPrimaryDestinationAccount(user).balance, 0.0
        )

        self.assertEqual(Account.GetPrimarySourceAccount(user).balance, 0.0)

    def test_balance(self):
        Transaction.objects.create(
            amount=50.0,
            source_account=self.user1_source,
            destination_account=self.user1_destination,
            is_settled=True,
        )

        self.assertEqual(self.user1_source.balance, 50.0)

    def test_transfer(self):
        Transaction.objects.create(
            amount=50.0,
            source_account=self.user1_source,
            destination_account=self.user1_destination,
            is_settled=True,
        )

        self.user1_source.transfer(50.0, self.user2_destination)
        self.assertEqual(self.user2_destination.balance, 50.0)

    def test_transfer_insufficient_balance(self):
        self.assertEqual(self.user1_source.balance, 0.0)

        with self.assertRaises(Account.InsufficientBalance):
            self.user1_source.transfer(50.0, self.user2_destination)



# coding: utf8

import logging

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import models
from django.dispatch import receiver

from model_utils.models import TimeStampedModel

log = logging.getLogger('accountant.models')


class Account(TimeStampedModel):
    comment = models.CharField(max_length=200, blank=True)
    is_primary_destination = models.BooleanField(default=True)
    is_primary_source = models.BooleanField(default=True)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='accounts')

    class InsufficientBalance(Exception):
        def __init__(self, balance, amount):
            self.balance = balance
            self.amount = amount

        def __str__(self):
            return str('%0.2f < %0.2f' % (self.balance, self.amount))

        def __unicode__(self):
            return '%0.2f < %0.2f' % (self.balance, self.amount)

    def __unicode__(self):
        if self.comment:
            return unicode('%s (%s)' % (self.user, self.comment))
        return unicode(self.user)

    @classmethod
    def GetMasterAccount(cls):
        """ Returns the site master account. """

        return cls.objects.get(pk=settings.MASTER_ACCOUNT_PK)

    @classmethod
    def GetPrimaryDestinationAccount(cls, user):
        """ Returns the given user's primary destination account. """

        try:
            return cls.objects.get(user=user, is_primary_destination=True)
        except cls.DoesNotExist:
            log.error(
                'Primary destination account for user: %s does not exist.'
                % user
            )

    @classmethod
    def GetPrimarySourceAccount(cls, user):
        """ Returns the given user's primary source account. """

        try:
            return cls.objects.get(user=user, is_primary_source=True)
        except cls.DoesNotExist:
            log.error(
                'Primary source account for user: %s does not exist.'
                % user
            )

    def get_balance(self, only_settled=True):
        """
        Sums the source transactions less than zero and the destination
        transactions greater than zero.
        """
        balance = sum(map(
            lambda t: t.amount, self.source_transactions.filter(
                amount__lt=0, is_settled=only_settled
            )
        ))

        balance += sum(map(
            lambda t: t.amount, self.destination_transactions.filter(
                amount__gt=0, is_settled=only_settled
            )
        ))

        return float(balance)
    balance = property(get_balance)

    def transfer(self, amount, account, comment=''):
        """
        Transfers the amount from this account to the given account if the
        account has enough funds.
        """
        if self.balance < amount:
            log.error(
                'Account %d does not have enough funds to cover the transfer. '
                'Balance: %0.2f, Amount Requested: %0.2f, Comment: %s' % (
                    self.pk, self.balance, amount, comment
                )
            )

            raise self.__class__.InsufficientBalance(self.balance, amount)

        Transaction.objects.create(
            amount=amount * -1,
            is_settled=True,
            source_account=self,
            destination_account=account,
            comment=comment,
        )

        destination_transaction = Transaction.objects.create(
            amount=amount,
            is_settled=True,
            source_account=self,
            destination_account=account,
            comment=comment,
        )

        return destination_transaction


@receiver(models.signals.post_save,
          sender=get_user_model(),
          dispatch_uid='create_user_account')
def create_user_account(sender, instance, created, **kwargs):
    if getattr(settings, 'ACCOUNTANT_CREATE_USER_ACCOUNT', True) and created:
        Account.objects.create(
            user=instance,
            is_primary_destination=True,
            is_primary_source=True,
        )


class Transaction(TimeStampedModel):
    amount = models.DecimalField(decimal_places=4, max_digits=9)
    comment = models.CharField(max_length=200, blank=True, default='')
    is_settled = models.BooleanField(default=False)

    source_account = models.ForeignKey(
        'accountant.Account',
        related_name='source_transactions'
    )

    destination_account = models.ForeignKey(
        'accountant.Account',
        related_name='destination_transactions'
    )

    def __unicode__(self):
        return unicode('%f %s (%d) %s -> %s' % (
            self.amount,
            self.reference_type,
            self.reference_id,
            self.source_account,
            self.destination_account
        ))
